function setup_functions()
	local cached_api_item_data = {}
	local function setup_functions_raw()
		local function_debug_output_enabled = false
		function print_debug(...)
			if function_debug_output_enabled then
				print(...)
			end
		end
		function enable_function_debug_output(newstate)
			if newstate == nil then
				newstate = true
			end
			function_debug_output_enabled = newstate
		end

		function status()
			return get_cached_function(get_status_info)
		end
		function inventory()
			return get_cached_function(get_inventory_counts)
		end
		function itemid_api_data(itemid)
			if not cached_api_item_data[itemid] then
				cached_api_item_data[itemid] = get_api_itemid_info(itemid)
			end
			return cached_api_item_data[itemid]
		end

		function buffslist()
			local tbl = {}
			for x in table.values(status().effects) do
				tbl[x[1]] = tonumber(x[2])
			end
			return tbl
		end

		function classid() return tonumber(status().class) end

		function playerid() return tonumber(status().playerid) end

		-- WARNING: Values can be out of date unless you load charpane.php. This is a KoL/CDM bug.

		function get_mainstat()
			local cid = classid()
			if cid == 1 or cid == 2 then
				return "Muscle"
			elseif cid == 3 or cid == 4 then
				return "Mysticality"
			elseif cid == 5 or cid == 6 then
				return "Moxie"
			else
				return "Muscle"
			end
		end

-- 		function level() return tonumber(status().level) end -- doesn't update when it should and was also just bugged before, KoL/CDM bug
		function level()
			if basemainstat() < 4 then
				return 1
			else
				return 1 + math.floor(math.sqrt(basemainstat() - 4))
			end
		end

		function playername() return status().name end
		function ascensions_count() return tonumber(status().ascensions) end
		function current_ascension_number() return ascensions_count() + 1 end
		function daysthisrun() return tonumber(status().daysthisrun) end
		function meat() return tonumber(status().meat) end
		function maxhp() return tonumber(status().maxhp) end
		function maxmp() return tonumber(status().maxmp) end
		-- Serverside API bug, HP/MP can be larger than MaxHP/MaxMP
		function hp() return math.min(tonumber(status().hp), maxhp()) end
		function mp() return math.min(tonumber(status().mp), maxmp()) end
		function turnsthisrun() return tonumber(status().turnsthisrun) end
		function familiarid() return tonumber(status().familiar) end
		function familiarpicture() return status().familiarpic end
		function buffedfamiliarweight() return tonumber(status().famlevel) end
		function buff(name) return buffslist()[name] ~= nil end
		function buffturns(name) return buffslist()[name] or 0 end
		function advs() return tonumber(status().adventures) end
		function buffedmuscle() return tonumber(status().muscle) end
		function buffedmysticality() return tonumber(status().mysticality) end
		function buffedmoxie() return tonumber(status().moxie) end
-- 		function basemuscle() return tonumber(status().basemuscle) end -- this don't update correctly
-- 		function basemysticality() return tonumber(status().basemysticality) end -- this don't update correctly
-- 		function basemoxie() return tonumber(status().basemoxie) end -- this don't update correctly
		function rawmuscle() return tonumber(status().rawmuscle) end
		function rawmysticality() return tonumber(status().rawmysticality) end
		function rawmoxie() return tonumber(status().rawmoxie) end
		function basemuscle()
			return math.floor(math.sqrt(rawmuscle()))
		end
		function basemysticality()
			return math.floor(math.sqrt(rawmysticality()))
		end
		function basemoxie()
			return math.floor(math.sqrt(rawmoxie()))
		end
		function buffedmainstat()
			local stats = {
				Muscle = buffedmuscle(),
				Mysticality = buffedmysticality(),
				Moxie = buffedmoxie(),
			}
			return stats[get_mainstat()]
		end
		function basemainstat()
			local stats = {
				Muscle = basemuscle(),
				Mysticality = basemysticality(),
				Moxie = basemoxie(),
			}
			return stats[get_mainstat()]
		end
		function rawmainstat()
			local stats = {
				Muscle = rawmuscle(),
				Mysticality = rawmysticality(),
				Moxie = rawmoxie(),
			}
			return stats[get_mainstat()]
		end
		function lastadventuredata() return status().lastadv end
		function ascensionpathid() return tonumber(status().path) end
		function ascensionpathname() return status().pathname end
		function moonsign() return status().sign end
		function moonsign_area()
			local areas = {
				Mongoose = "Degrassi Knoll",
				Wallaby = "Degrassi Knoll",
				Vole = "Degrassi Knoll",
				Platypus = "Little Canadia",
				Opossum = "Little Canadia",
				Marmot = "Little Canadia",
				Wombat = "Gnomish Gnomad Camp",
				Blender = "Gnomish Gnomad Camp",
				Packrat = "Gnomish Gnomad Camp",
			}
			return areas[moonsign()]
		end
		function equipment()
			local eq = {}
			for a, b in pairs(status().equipment) do
				eq[a] = tonumber(b)
			end
			eq.fakehands = nil -- Work around API misfeature - this is not an itemid
			return eq
		end
		function fullness() return tonumber(status().full) end
		function drunkenness() return tonumber(status().drunk) end
		function spleen() return tonumber(get_daily_counter("spleen")) end
		function spleen_display_string()
			local spleen = tostring(get_daily_counter("spleen"))
			if day["unknown spleen"] == "yes" then
				spleen = spleen .. "+?"
			end
			return spleen
		end
		function ascensionstatus()
			if tonumber(status().freedralph) == 1 then
				return "Aftercore"
			elseif tonumber(status().casual) == 1 then
				return "Aftercore"
			elseif tonumber(status().hardcore) == 1 then
				return "Hardcore"
			elseif tonumber(status().roninleft) > 0 then
				return "Softcore"
			else
				return "Aftercore"
			end
		end
		function in_aftercore()
			return ascensionstatus() == "Aftercore"
		end
		function mcd() return tonumber(status().mcd) end
		function applied_scratchnsniff_stickers() return status().stickers end
		function api_flag_config() return status().flag_config end

		function substats_for_level(x)
			if x == 1 then
				return 9
			else
				local needstat = (x - 1) * (x - 1) + 4
				local needsubstat = needstat * needstat
				return needsubstat
			end
		end

		function level_progress()
			local basesubstat = substats_for_level(level())
			local nextsubstat = substats_for_level(level() + 1)
			local need = nextsubstat - basesubstat
			local have = rawmainstat() - basesubstat
-- 			print("progress", basesubstat, nextsubstat, need, have)
			return have / need, have, need
		end

		local function get_cached_equip_counts()
			return get_cached_item("cached_equip_counts", function()
				local tbl = {}
				for x in table.values(equipment()) do
					tbl[x] = (tbl[x] or 0) + 1
				end
				return tbl
			end)
		end

		function have_inventory(name)
			if inventory()[get_itemid(name)] then
				return true
			else
				return false
			end
		end

		function have_equipped_itemid(id)
			return get_cached_equip_counts()[id] ~= nil
		end

		function count_equipped_itemid(id)
			return get_cached_equip_counts()[id] or 0
		end

		function have_itemid(id)
			if inventory()[id] then
				return true
			else
				return have_equipped_itemid(id)
			end
		end

		function have(name)
			return have_itemid(get_itemid(name))
		end

		function have_equipped(name)
			return have_equipped_itemid(get_itemid(name))
		end

		function count_equipped(name)
			return count_equipped_itemid(get_itemid(name))
		end

		function count_itemid(id)
			return (inventory()[id] or 0) + count_equipped_itemid(id)
		end

		function count(name)
			return count_itemid(get_itemid(name))
		end

		function count_inventory(name)
			return inventory()[get_itemid(name)] or 0
		end

		function use_hottub()
			return async_get_page("/clan_viplounge.php", { action = "hottub" })
		end

		function get_wand_data()
			local wands = { "aluminum wand", "ebony wand", "hexagonal wand", "marble wand", "pine wand" }
			for _, x in ipairs(wands) do
				if have(x) then
					local itemid = get_itemid(x)
					local pt = get_page("/wand.php", { whichwand = itemid })
					if pt:contains("Zap an item") then
						if pt:contains(x) or pt:contains("Your wand ") or pt:contains("feels warm") or pt:contains("be careful") then
							return { name = x, itemid = itemid, heat = 1 }
						else
							return { name = x, itemid = itemid, heat = 0 }
						end
					end
				end
			end
			return nil
		end

		function meatpaste_items(a, b) -- TODO: can this be done without requiring up-to-date status?
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			if moonsign_area() == "Degrassi Knoll" then
				return async_post_page("/knoll.php", { action = "combine", pwd = pwd, item1 = get_itemid(a), item2 = get_itemid(b), quantity = 1, ajax = 1 })
			else
				if not have("meat paste") then
					async_post_page("/craft.php", { pwd = pwd, action = "makepaste", qty = 1, ajax = 1, whichitem = get_itemid("meat paste") })
				end
				return async_post_page("/craft.php", { mode = "combine", pwd = pwd, action = "craft", a = get_itemid(a), b = get_itemid(b), qty = 1, ajax = 1 })
			end
		end

		function cook_items(a, b)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_post_page("/craft.php", { mode = "cook", pwd = pwd, action = "craft", a = get_itemid(a), b = get_itemid(b), qty = 1, ajax = 1 })
		end

		function mix_items(a, b)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_post_page("/craft.php", { mode = "cocktail", pwd = pwd, action = "craft", a = get_itemid(a), b = get_itemid(b), qty = 1, ajax = 1 })
		end

		function smith_items(a, b)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_post_page("/knoll.php", { action = "smith", pwd = pwd, item1 = get_itemid(a), item2 = get_itemid(b), quantity = 1, ajax = 1 })
		end

		function buy_item(name, whichstore, amount)
			print_debug("  buying item", name, amount or "")
			local pwd = session.pwd -- inserting pwd, boo!
-- 			print("buying item", name, amount or "")
			amount = amount or 1
			return async_get_page("/store.php", { phash = pwd, buying = 1, whichitem = get_itemid(name), howmany = amount, whichstore = whichstore, ajax = 1, action = "buyitem" })
		end

		function equip_itemid(id, slot)
			print_debug("  equipping", slot, get_item_data_by_id(tonumber(id)).name)
			local pwd = session.pwd -- inserting pwd, boo!
			local tbl = { pwd = pwd, action = "equip", whichitem = id, ajax = 1 }
			if slot == "dualwield" then
				tbl.action = "dualwield"
			elseif slot == "offhand" then
				tbl.action = "dualwield"
				local f = get_page("/inv_equip.php", tbl)
				if equipment().offhand ~= id then
					tbl.action = "equip"
					tbl.slot = tostring(slot)
					return async_get_page("/inv_equip.php", tbl)
				else
					return f()
				end
			elseif slot == "acc1" then
				tbl.slot = "1"
			elseif slot == "acc2" then
				tbl.slot = "2"
			elseif slot == "acc3" then
				tbl.slot = "3"
			elseif slot == "familiarequip" then
				tbl.action = "hatrack"
				local f = async_get_page("/inv_equip.php", tbl)
				if equipment().familiarequip ~= id then
					tbl.action = "equip"
					tbl.slot = tostring(slot)
					return async_get_page("/inv_equip.php", tbl)
				else
					return f()
				end
			elseif slot then
				tbl.slot = tostring(slot)
			end
			return async_get_page("/inv_equip.php", tbl)
		end

		function equip_item(name, slot)
			return equip_itemid(get_itemid(name), slot)
		end

		function unequip_slot(name)
			print_debug("  unequipping ", name)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_get_page("/inv_equip.php", { pwd = pwd, action = "unequip", type = name, ajax = "1" })
		end

		function set_mcd(amount)
			local pwd = session.pwd -- inserting pwd, boo!
			-- TODO: non-muscle signs
			return async_get_page("/inv_use.php", { pwd = pwd, whichitem = get_itemid("detuned radio"), ajax = 1, tuneradio = amount })
		end

		function set_equipment(tbl)
-- 			print("setting equipment to", tbl)
			local eq = equipment()
			for a, b in pairs(eq) do
-- 				print("checking", b, "vs[", a, "]", tbl[a])
				if b ~= tbl[a] then
					unequip_slot(a)
-- 					print("unequipping slot", a)
				end
			end
			local eq = equipment()
			for a, b in pairs(tbl) do
				if eq[a] ~= b then
-- 					print("equip_itemid", b, a)
					equip_itemid(b, a)
				end
			end
			local eq = equipment()
			local function getnamedesc(id)
				if tonumber(id) then
					local d = get_item_data_by_id(tonumber(id))
					if d and d.name then
						return (d.name .. " (itemid:" .. tonumber(id) .. ")")
					else
						return ("itemid:" .. tonumber(id))
					end
				elseif id then
					return id
				else
					return "nil"
				end
			end
			for a, b in pairs(eq) do
				if b ~= tbl[a] then
					error("Wearing " .. tostring(a) .. ":" .. tostring(b) .. " after trying to wear " .. getnamedesc(tbl[a]))
				end
			end
			for a, b in pairs(tbl) do
				if eq[a] ~= b then
					error("Wearing " .. tostring(a) .. ":" .. tostring(eq[a]) .. " after trying to wear " .. getnamedesc(b))
				end
			end
		end

		function use_itemid(id, amount)
			local pwd = session.pwd -- inserting pwd, boo!
			if amount then
				return async_get_page("/multiuse.php", { pwd = pwd, whichitem = id, ajax = 1, quantity = amount, action = "useitem" })
			else
				return async_get_page("/inv_use.php", { pwd = pwd, whichitem = id, ajax = 1 })
			end
		end

		function use_itemid_noajax(id, amount)
			local pwd = session.pwd -- inserting pwd, boo!
			if amount then
				return async_get_page("/multiuse.php", { pwd = pwd, whichitem = id, quantity = amount, action = "useitem" })
			else
				return async_get_page("/inv_use.php", { pwd = pwd, whichitem = id })
			end
		end

		function use_item(name, amount)
			print_debug("  using item", name, amount or "")
			return use_itemid(get_itemid(name), amount)
		end

		function use_item_noajax(name, amount)
			return use_itemid_noajax(get_itemid(name), amount)
		end

		function eat_item(name)
			print_debug("  eating item", name)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_get_page("/inv_eat.php", { pwd = pwd, which = 1, whichitem = get_itemid(name), ajax = 1 })
		end

		function drink_itemid(id)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_get_page("/inv_booze.php", { pwd = pwd, whichitem = id, ajax = 1 })
		end

		function drink_item(name)
			print_debug("  drinking item", name)
			drink_itemid(get_itemid(name))
		end

		function pull_itemids(xs)
			local pwd = session.pwd -- inserting pwd, boo!
			local tbl = { pwd = pwd, action = "pull" }
			for nr, id in pairs(xs) do
				tbl["howmany" .. tostring(nr)] = 1
				tbl["whichitem" .. tostring(nr)] = id
			end
			-- TODO!: split up more than 11
			return async_post_page("/storage.php", tbl)
		end

		function freepull_itemid(id)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_post_page("/storage.php", { action = "pull", pwd = pwd, howmany1 = 1, whichitem1 = id })
		end

		function closet_itemid(id)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_get_page("/inventory.php", { action = "closetpush", pwd = pwd, qty = 1, whichitem = id, ajax = 1 })
		end

		function uncloset_itemid(id)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_get_page("/inventory.php", { action = "closetpull", pwd = pwd, qty = 1, whichitem = id, ajax = 1 })
		end
		
		function sell_itemid(id, amount)
			local pwd = session.pwd -- inserting pwd, boo!
			return async_get_page("/sellstuff.php", { action = "sell", ajax = 1, type = "quant", ["whichitem[]"] = id, howmany = amount or 1, pwd = pwd })
		end

		function cast_skillid(skillid, quantity, targetid)
			local pwd = session.pwd -- inserting pwd, boo!
			targetid = targetid or playerid()
			assert(targetid and targetid ~= "")
			local tbl = { whichskill = skillid, ajax = 1, action = "Skillz", pwd = pwd, targetplayer = targetid, quantity = quantity }
			return async_get_page("/skills.php", tbl)
		end

		function switch_familiarid(id)
			return async_get_page("/familiar.php", { action = "newfam", ajax = 1, newfam = id })
		end

		function handle_adventure_result(pt, url, zoneid, macro, noncombatchoices, specialnoncombatfunction)
			if url:contains("/fight.php") then -- TODO: hmmmm? -- and url:match("ireallymeanit") then
				local advagain = nil
				if pt:contains([[>You win the fight!<!--WINWINWIN--><]]) then
					advagain = true
				elseif pt:contains([[state['fightover'] = true;]]) or true then -- HACK: doesn't get set with combat bar disabled
					if pt:contains("You lose.") then
						advagain = false
					elseif pt:contains([[<a href="adventure.php?snarfblat=]]..zoneid..[[">Adventure Again]]) then
						advagain = true
					end
				end
				if advagain == nil then
					if macro then
						local pt, url = post_page("/fight.php", { action = "macro", macrotext = macro })
--~ 						print("recurse with macro")
						return handle_adventure_result(pt, url, zoneid, nil, noncombatchoices, specialnoncombatfunction)
					else
						print("fight.php unhandled url", url)
					end
				end
-- 				print("return1 p u a", pt:len(), url, advagain)
				return pt, url, advagain
			elseif url:contains("/choice.php") then
				local advagain = nil
				local adventure_title = pt:match([[<tr><td style="color: white;" align=center bgcolor=blue.-><b>([^<]*)</b></td></tr>]])
				if adventure_title == "Results:" then
					if pt:contains([[<a href="adventure.php?snarfblat=]]..zoneid..[[">Adventure Again]]) then
						advagain = true
					end
-- 					print("return2 p u a", pt:len(), url, advagain)
					return pt, url, advagain
				end
				local choice_adventure_number = tonumber(pt:match([[<input type=hidden name=whichchoice value=([0-9]+)>]]))
		--~ 		print("choice", adventure_title, choice_adventure_number)
				local pickchoice = nil
				local optname = nil
				if specialnoncombatfunction then
					optname, pickchoice = specialnoncombatfunction(adventure_title, choice_adventure_number, pt)
				else
					optname = noncombatchoices[adventure_title]
				end
				if optname and not pickchoice then
					for nr, title in pt:gmatch([[<input type=hidden name=option value=([0-9])><input class=button type=submit value="([^>]+)">]]) do
			--~ 			print("opt", nr, title)
						if title == optname then
							pickchoice = tonumber(nr)
						end
					end
				end
				if optname and not pickchoice then
					print("Warning: option " .. tostring(optname) .. " not found for " .. tostring(adventure_title) .. ".")
				end
				if pickchoice then
					local pwd = session.pwd -- inserting pwd, boo!
					local pt, url = post_page("/choice.php", { pwd = pwd, whichchoice = choice_adventure_number, option = pickchoice })
-- 					print("choice ->", url)
					return handle_adventure_result(pt, url, zoneid, macro, noncombatchoices, specialnoncombatfunction)
				else
					print("choice", adventure_title, choice_adventure_number)
					for nr, title in pt:gmatch([[<input type=hidden name=option value=([0-9])><input class=button type=submit value="([^>]+)">]]) do
						print("opt", nr, title)
					end
-- 					print("return3 p u a", pt:len(), url, advagain)
					return pt, url, false
				end
			else
				local advagain = false
				if pt:contains([[<a href="adventure.php?snarfblat=]]..zoneid..[[">Adventure Again]]) then
					advagain = true
-- 				else
-- 					print("non-fight non-choice unhandled url", url)
				end
-- 				print("return4 p u a", pt:len(), url, advagain)
				return pt, url, advagain
			end
		end

		function autoadventure(tbl)
			if not tbl.ignorewarnings and character["enable adventure warnings"] ~= "no" then
				local foo = { kolproxy_log_time_interval("check adv warnings", function()
					local warn_tbl = get_raw_adventure_warnings()
	-- 				print("warn_tbl is", warn_tbl)
					for f in table.values(warn_tbl) do
						local message, warningid, custommsg = f(tbl.zoneid)
						if message then
							print("advwarn?", f, message, warningid)
							local x, y = intercept_warning { message = message, id = warningid, customdisablemsg = custommsg, norepeat = true }
							if x then
-- 								print("  warn!", x, y, "...")
								return x, y, false
							end
						end
					end
				end) }
				if foo[1] then
					return unpack(foo)
				end
			end
			fight.zone = tbl.zoneid
			session["adventure.lastzone"] = tbl.zoneid
			local pt, url = post_page("/adventure.php", { snarfblat = tbl.zoneid })
			return handle_adventure_result(pt, url, tbl.zoneid, tbl.macro, tbl.noncombatchoices or {}, tbl.specialnoncombatfunction)
		end

		function raw_retrieve_skills()
			local cs = get_page("/charsheet.php")
			local skills_text = cs:match("<p>Skills:</b>.-(<a onClick.-)</td>")
			if not skills_text then
				return nil
			end
			local skills = {}
			for s in skills_text:gmatch("<a onClick.->(.-)</a>") do
				skills[s] = true
			end
			return skills
		end

		function get_skills()
			return get_cached_item("cached_get_skills", function()
				-- TODO: Cache in a variable if merging Lua states
				local cached_skills = session["cached skills"]
				if not cached_skills then
					cached_skills = raw_retrieve_skills()
					session["cached skills"] = cached_skills
				end
				return cached_skills
			end)
		end
		function clear_cached_skills()
			session["cached skills"] = nil
		end
		function have_skill(name)
			if character["run automation scripts"] == "no" then
				return nil
			end
			local skills = get_skills()
			if skills then
				return skills[name] ~= nil
			end
		end

		function retrieve_trailed_monster()
			local effectpt = get_page("/desc_effect.php", { whicheffect = "91635be2834f8a07c8ff9e3b47d2e43a" })
			local trailed = effectpt:match([[And by "wabbit" I mean "(.-)%."]])
			return trailed
		end
	end

	local s_f_env = {}
	setmetatable(s_f_env, { __index = _G, __newindex = _G })
	setfenv(setup_functions_raw, s_f_env)
	setup_functions_raw()
	setmetatable(s_f_env, { __index = _G,
	__newindex = function(t, k, v)
		print("warning: s_f_env __newindex", t, k, v)
		rawset(t, k, v)
	end })
end
